#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""The version number is based on semantic versioning.

References
- https://semver.org/
- https://www.python.org/dev/peps/pep-0440/

This file is autogenerated, do not modify by hand.
"""
__version__ = "1.7.2"
COMMIT = "580f118748fc89458692ffa13b9dc0b78c0db5ae"
MAJOR = 1
MINOR = 7
PATCH = 2
