#
# Copyright (C) 2020 Arm Mbed. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Scripts and utilities used by the CI pipeline."""
from mbed_tools_ci_scripts._version import __version__
